from microcosm.api import defaults, typed
from microcosm.config.types import boolean
from microcosm.errors import NotBoundError

from microcosm_pubsub.result import MessageHandlingResult, MessageHandlingResultType


@defaults(
    enabled=typed(boolean, default_value=True)
)
class PubSubSendMetrics:
    """
    Send metrics relating to a single MessageHandlingResult

    """

    def __init__(self, graph):
        self.metrics = self.get_metrics(graph)
        self.enabled = bool(
            self.metrics
            and self.metrics.host != "localhost"
            and graph.config.pubsub_send_metrics.enabled
        )

    def get_metrics(self, graph):
        """
        Fetch the metrics client from the graph.

        Metrics will be disabled if the not configured.

        """
        try:
            return graph.metrics
        except NotBoundError:
            return None

    def __call__(self, result: MessageHandlingResult):
        """
        Send metrics if enabled.

        """
        if not self.enabled:
            return

        if result.result == MessageHandlingResultType.IGNORED:
            return

        tags = [
            "source:microcosm-pubsub",
            f"result:{result.result}",
            f"media-type:{result.media_type}",
        ]
        self.metrics.histogram(
            "message",
            result.elapsed_time,
            tags=tags,
        )

        if result.handle_start_time:
            self.metrics.histogram(
                "message_handle_start",
                result.handle_start_time,
                tags=tags,
            )


@defaults(
    enabled=typed(boolean, default_value=True)
)
class PubSubSendBatchMetrics:
    """
    Send metrics relating to a batch of handled messages

    """

    def __init__(self, graph):
        self.metrics = self.get_metrics(graph)
        self.enabled = bool(
            self.metrics
            and self.metrics.host != "localhost"
            and graph.config.pubsub_send_metrics.enabled
        )

    def get_metrics(self, graph):
        """
        Fetch the metrics client from the graph.

        Metrics will be disabled if the not configured.

        """
        try:
            return graph.metrics
        except NotBoundError:
            return None

    def __call__(self, elapsed_time: float, message_batch_size: int):
        """
        Send metrics if enabled, and if the batch processed contains at least one
        non-ignored message

        """
        if not self.enabled:
            return

        if not message_batch_size > 0:
            return

        tags = [
            "source:microcosm-pubsub",
        ]

        self.metrics.histogram(
            "message_batch",
            elapsed_time,
            tags=tags,
        )

        self.metrics.histogram(
            "message_batch_size",
            message_batch_size,
            tags=tags,
        )


@defaults(
    enabled=typed(boolean, default_value=True)
)
class PubSubProducerMetrics:
    """
    Send metrics regarding the producer

    """

    def __init__(self, graph):
        self.metrics = self.get_metrics(graph)
        self.enabled = bool(
            self.metrics
            and self.metrics.host != "localhost"
            and graph.config.pubsub_send_metrics.enabled
        )

    def get_metrics(self, graph):
        """
        Fetch the metrics client from the graph.

        Metrics will be disabled if the not configured.

        """
        try:
            return graph.metrics
        except NotBoundError:
            return None

    def __call__(self, elapsed_time: float, publish_result: bool, **kwargs):
        """
        Send metrics for how long it takes to produce a message

        """
        if not self.enabled:
            return

        tags = [
            "source:microcosm-pubsub",
            f"publish_result:{publish_result}",
            f"media_type:{kwargs['media_type']}",
        ]

        self.metrics.histogram(
            "sns_producer",
            elapsed_time,
            tags=tags,
        )



"""
PublishBatchMessage handler.

"""

from microcosm.api import binding
from microcosm_logging.decorators import logger

from microcosm_pubsub.conventions import created
from microcosm_pubsub.decorators import handles
from microcosm_pubsub.producer import PubsubMessage, SNSProducer


@binding("publish_message_batch")
@handles(created("BatchMessage"))
@logger
class PublishBatchMessage:

    def __init__(self, graph):
        self.sns_producer: SNSProducer = graph.sns_producer

    def __call__(self, message):
        messages = message["messages"]
        for message in messages:
            pubsub_message = PubsubMessage(
                media_type=message["media_type"],
                message=message["message"],
                message_attributes=message["message_attributes"],
                opaque_data=message["opaque_data"],
                topic_arn=message["topic_arn"],
            )

            self.sns_producer.publish_message(pubsub_message)



from collections.abc import MutableMapping
from itertools import chain


CONTEXT = "context"


class SafeContext(MutableMapping):
    """
    Dictionary that raises on overwriting keys.

    """
    def __init__(self, *args, **kwargs):
        self.store = dict(*args, **kwargs)

    def __getitem__(self, key):
        # enable context self-references
        if key == CONTEXT:
            return self
        return self.store[key]

    def __setitem__(self, key, value):
        # do not allow overwrite of context self-references
        if key == CONTEXT:
            raise ValueError("May not reassign 'context' key")

        if key in self.store:
            raise ValueError(f"Key '{key}' already set")

        self.store[key] = value

    def assign(self, key, value):
        # do not allow overwrite of context self-references
        if key == CONTEXT:
            raise ValueError("May not reassign 'context' key")

        # Skip validation
        self.store[key] = value
        return self

    def __delitem__(self, key):
        del self.store[key]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __getattr__(self, key):
        return self[key]

    def local(self, *args, **kwargs):
        """
        Create a locally scoped child context.

        """
        return ScopedSafeContext(self, *args, **kwargs)


class ScopedSafeContext(SafeContext):
    """
    SafeContext that delegates reads to a parent context.

    """
    def __init__(self, parent, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent = parent

    def __getitem__(self, key):
        # enable context self-references
        if key == CONTEXT:
            return self

        try:
            return self.store[key]
        except KeyError:
            return self.parent[key]

    def __setitem__(self, key, value):
        if key in self.parent:
            raise ValueError(f"Key '{key}' already set")
        super().__setitem__(key, value)

    def __iter__(self):
        return chain(iter(self.store), iter(self.parent))

    def __len__(self):
        return len(self.store) + len(self.parent)



"""
Uri Handler base classes.

"""
from abc import ABCMeta
from re import search

from inflection import titleize
from microcosm.errors import LockedGraphError, NotBoundError
from requests import codes, get

from microcosm_pubsub.constants import DEFAULT_RESOURCE_CACHE_TTL
from microcosm_pubsub.conventions.lifecycle import LifecycleChange
from microcosm_pubsub.errors import Nack


def resource_cache_whitelist_callable(media_type, uri):
    """
    Default resource cache whitelist callable implementation.

    Only whitelists under the following conditions:
    * verb is created
    * resource is an event

    """
    if not media_type:
        # Nb. likely to happen with mocks and tests.
        return False

    return all((
        search(r"/[a-z]+_event/", uri),
        search(r".{}.".format(LifecycleChange.Created), media_type),
    ))


class URIHandler(metaclass=ABCMeta):
    """
    Base handler for URI-driven events.

    As a general rule, we want PubSub events to convey the URI of a resource that was created
    (because resources are ideally immutable state). In this case, we want asynchronous workers
    to query the existing URI to get more information (and to handle race conditions where the
    message was delivered before the resource was committed.)

    There are five expected outcomes for this handler:
      - Raising an error (e.g. a bug)
      - Skipping the handlers (because the pubsub message carried enough information to bypass processing)
      - Handling the message after fetching the resource by URI
      - Ignore the message after fetching the resource by URI
      - Raising a nack (e.g. because the resource was not committed yet)

    The middle three cases are all handled here with the expectation that we produce *one* INFO-level
    log per message processed (unless an error/nack is raised).

    """
    def __init__(
        self,
        graph,
        retry_nack_timeout=1,
        resource_nack_timeout=1,
        resource_cache_enabled=True,
        resource_cache_ttl=DEFAULT_RESOURCE_CACHE_TTL,
        resource_cache_whitelist_callable=resource_cache_whitelist_callable,
    ):
        self.opaque = graph.opaque
        self.retry_nack_timeout = retry_nack_timeout
        self.resource_nack_timeout = resource_nack_timeout
        self.resource_cache_ttl = resource_cache_ttl
        self.resource_cache = self.get_resource_cache(graph) if resource_cache_enabled else None
        self.resource_cache_whitelist_callable = resource_cache_whitelist_callable

    @property
    def nack_timeout(self):
        """Deprecated, use retry_nack_timeout"""
        return self.retry_nack_timeout

    @property
    def name(self):
        return titleize(self.__class__.__name__)

    @property
    def nack_if_not_found(self):
        return True

    @property
    def resource_type(self):
        return dict

    def get_resource_cache(self, graph):
        try:
            return graph.resource_cache
        except (LockedGraphError, NotBoundError):
            # Nb. if resource cache is globally disabled, will not be bound
            return None

    def __call__(self, message):
        uri = message["uri"]
        self.on_call(message, uri)

        skip_reason = self.get_reason_to_skip(message, uri)
        if skip_reason is not None:
            self.on_skip(message, uri, skip_reason)
            return False

        resource = self.convert_resource(
            self.get_resource(message, uri),
        )

        if self.handle(message, uri, resource):
            self.on_handle(message, uri, resource)
            return True
        else:
            self.on_ignore(message, uri, resource)
            return False

    def on_call(self, message, uri):
        self.logger.debug(
            "Starting {handler}",
            extra=dict(
                handler=self.name,
                uri=uri,
            ),
        )

    def on_skip(self, message, uri, reason):
        self.logger.info(
            "Skipping {handler} because {reason}",
            extra=dict(
                handler=self.name,
                reason=reason,
                uri=uri,
            ),
        )

    def on_handle(self, message, uri, resource):
        self.logger.debug(
            "Handled {handler}",
            extra=dict(
                handler=self.name,
                uri=uri,
            ),
        )

    def on_ignore(self, message, uri, resource):
        self.logger.info(
            "Ignored {handler}",
            extra=dict(
                handler=self.name,
                uri=uri,
            ),
        )

    def get_reason_to_skip(self, message, uri):
        """
        Some messages carry enough context that we can avoid resolving the URI entirely.

        """
        return None

    def validate_changed_field(self, message, resource):
        if message.get('field_name') and self.nack_if_not_found:
            field_name = message["field_name"]
            new_value = message["new_value"]
            if resource.get(field_name) != new_value:
                raise Nack(self.resource_nack_timeout)

    def get_resource(self, message, uri):
        """
        Mock-friendly URI getter.

        Passes message context.

        """
        if self.resource_cache and self.resource_cache_whitelist_callable(
            media_type=message.get("mediaType"),
            uri=uri
        ):
            response = self.resource_cache.get(uri)
            if response:
                return response

        headers = self.get_headers(message)
        response = get(uri, headers=headers)
        if response.status_code == codes.not_found and self.nack_if_not_found:
            raise Nack(self.resource_nack_timeout)
        response.raise_for_status()
        response_json = response.json()

        self.validate_changed_field(message, response_json)

        if self.resource_cache and self.resource_cache_whitelist_callable(
            media_type=message.get("mediaType"),
            uri=uri,
        ):
            self.resource_cache.set(uri, response_json, ttl=self.resource_cache_ttl)

        return response_json

    def get_headers(self, message):
        """
        Generate headers to pass to downstream services.

        """
        return self.opaque.as_dict()

    def convert_resource(self, resource):
        if isinstance(self.resource_type, type) and isinstance(resource, self.resource_type):
            return resource
        return self.resource_type(**resource)

    def handle(self, message, uri, resource):
        return True



"""
Syntactic sugar using existing primitives defined

"""
from microcosm_pubsub.chain import Chain
from microcosm_pubsub.chain.context import ScopedSafeContext


class LocalCallWrapper:
    """
    Syntactic sugar making it easy to call the same function multiple times in a chain
    Usage:
    Chain(
       ...,
       call(a_function).with_args(a_function_arg="an_existing_context_key").as_("name_of_the_result_in_context"),
       call(a_function).with_args(a_function_arg="some_other_key").as_("another_name"),
    )

    """
    def __init__(self, func):
        self.func = func
        self.chain = Chain(func)
        self.kwargs = dict()
        self.result_name = None
        self.local_context = None

    def with_args(self, **kwargs):
        self.kwargs = kwargs
        return self

    def as_(self, name):
        self.result_name = name
        return self

    def _build_local_kwargs(self, parent_context):
        local_kwargs = dict()

        for local_name, mapped_value in self.kwargs.items():
            if local_name in parent_context:
                raise ValueError(f"Argument `{local_name}` for local call overshadows existing context key")

            if isinstance(mapped_value, str) and mapped_value in parent_context:
                # mapped value points to parent context
                local_kwargs[local_name] = parent_context[mapped_value]
            else:
                # mapped value is a constant
                local_kwargs[local_name] = mapped_value

        return local_kwargs

    def __call__(self, context):
        local_kwargs = self._build_local_kwargs(context)
        self.local_context = ScopedSafeContext(context, **local_kwargs)
        result = self.chain(self.local_context)
        context[self.result_name] = result
        return result


def call(func):
    return LocalCallWrapper(func)



"""
Message producer.

"""
from collections import defaultdict
from dataclasses import dataclass
from distutils.util import strtobool
from functools import wraps
from logging import Logger
from time import time
from typing import Dict, List

from boto3 import Session
from botocore.client import Config
from microcosm.api import defaults, typed
from microcosm.errors import NotBoundError
from microcosm_logging.decorators import logger
from microcosm_logging.timing import elapsed_time

from microcosm_pubsub.batch import MessageBatchSchema
from microcosm_pubsub.constants import PUBLISHED_KEY
from microcosm_pubsub.conventions.naming import make_media_type
from microcosm_pubsub.errors import TopicNotDefinedError


@dataclass
class PubsubMessage:
    """
    Container class encapsulating all of the necessary components for
    publishing a given message

    """
    media_type: str
    message: str
    message_attributes: Dict[str, Dict[str, str]]
    opaque_data: dict
    topic_arn: str


@logger
class SNSProducer:
    """
    Produces messages to SNS topics.

    """
    logger: Logger

    def __init__(
        self,
        opaque,
        pubsub_message_schema_registry,
        sns_client,
        sns_topic_arns,
        skip,
        deferred_batch_size,
        pubsub_producer_metrics
    ):
        self.opaque = opaque
        self.pubsub_message_schema_registry = pubsub_message_schema_registry
        self.sns_client = sns_client
        self.sns_topic_arns = sns_topic_arns
        self.skip = skip
        self.deferred_batch_size = deferred_batch_size
        self.pubsub_producer_metrics = pubsub_producer_metrics

    def produce(self, media_type, dct=None, uri=None, **kwargs):
        """
        Produce a message.

        :returns: the message id

        """

        if self.skip:
            return
        pubsub_message = self.create_message(media_type, dct, uri, **kwargs)
        return self.publish_message(pubsub_message)

    def create_message(self, media_type, dct, uri=None, opaque_data=None, **kwargs) -> PubsubMessage:
        if opaque_data is None:
            opaque_data = dict()

        if self.opaque is not None:
            opaque_data.update(self.opaque.as_dict())

        opaque_data[PUBLISHED_KEY] = str(time())

        topic_arn = self.choose_topic_arn(media_type)

        message_attributes = self.choose_message_attributes(media_type)

        message = self.pubsub_message_schema_registry.find(media_type).encode(
            dct,
            opaque_data=opaque_data,
            media_type=media_type,
            uri=uri,
            **kwargs
        )
        return PubsubMessage(
            media_type=media_type,
            message=message,
            message_attributes=message_attributes,
            opaque_data=opaque_data,
            topic_arn=topic_arn,
        )

    def publish_message(self, pubsub_message: PubsubMessage):
        extra = dict(
            media_type=pubsub_message.media_type,
            **pubsub_message.opaque_data
        )
        self.logger.debug("Publishing message with media type {media_type}", extra=extra)
        publish_result = "SUCCESS"
        publish_exception = None

        with elapsed_time(extra):
            try:
                result = self.sns_client.publish(
                    TopicArn=pubsub_message.topic_arn,
                    Message=pubsub_message.message,
                    MessageAttributes=pubsub_message.message_attributes,
                )
            except Exception as e:
                publish_exception = e
                publish_result = "FAILURE"

        self.pubsub_producer_metrics(
            publish_result=publish_result,
            **extra
        )

        if publish_exception:
            raise Exception(f"Could not publish message, SNS producer error: {publish_exception}")

        self.logger.debug("Published message with media type {media_type}", extra=extra)

        # This is another layer of protection for the message send
        # If result is not defined, it will exit with an error here
        return result["MessageId"]

    def choose_topic_arn(self, media_type):
        """
        Choose a topic for this type of message.

        """
        try:
            topic_arn = self.sns_topic_arns[media_type]
        except KeyError:
            topic_arn = None

        if topic_arn is None:
            raise TopicNotDefinedError("No topic arn was registered for messages of type: {}".format(
                media_type,
            ))
        return topic_arn

    def choose_message_attributes(self, media_type: str) -> Dict[str, Dict[str, str]]:
        """
        Choose message attributes for this message

        """
        return {
            "media_type": {
                "DataType": "String",
                "StringValue": media_type,
            }
        }


class DeferredProducer:
    """
    A context manager to defer message production until the end of a block.

    """
    def __init__(self, producer):
        self.producer = producer
        self.messages = []

    def produce(self, media_type, dct=None, **kwargs):
        if self.producer.skip:
            return

        pubsub_message = self.producer.create_message(media_type, dct, **kwargs)
        self.messages.append(pubsub_message)

    def __enter__(self):
        self.messages = []
        return self

    def __exit__(self, type, value, traceback):
        if type is not None:
            return

        for pubsub_message in self.messages:
            self.producer.publish_message(pubsub_message)


class DeferredBatchProducer(DeferredProducer):

    def generate_message_batches(self, messages, deferred_batch_size):
        for i in range(0, len(messages), deferred_batch_size):
            yield messages[i:i + deferred_batch_size]

    def construct_batch_pubsub_message(self, message_batch: List[PubsubMessage]):
        return [
            dict(
                media_type=pubsub_message.media_type,
                message=pubsub_message.message,
                message_attributes=pubsub_message.message_attributes,
                opaque_data=pubsub_message.opaque_data,
                topic_arn=pubsub_message.topic_arn,
            )
            for pubsub_message in message_batch
        ]

    def __exit__(self, type, value, traceback):
        if type is not None:
            return

        for message_batch in self.generate_message_batches(self.messages, self.producer.deferred_batch_size):
            if len(message_batch) > 1:
                self.producer.produce(
                    MessageBatchSchema.MEDIA_TYPE,
                    messages=self.construct_batch_pubsub_message(message_batch),
                )
            elif len(message_batch) == 1:
                self.producer.publish_message(message_batch[0])


def deferred(component, key="sns_producer"):
    """
    A decorator to defer message production until after the decorated function has completed

    """
    graph = component.graph
    sns_producer = getattr(graph, key)

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                deferred_producer = DeferredProducer(sns_producer)
                setattr(component, key, deferred_producer)
                with deferred_producer:
                    return func(*args, **kwargs)
            finally:
                setattr(component, key, sns_producer)
        return wrapper
    return decorator


def deferred_batch(component, key="sns_producer"):
    """
    A decorator to defer batch message production until after the decorated function has completed

    """
    graph = component.graph
    sns_producer = getattr(graph, key)

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                deferred_producer = DeferredBatchProducer(sns_producer)
                setattr(component, key, deferred_producer)
                with deferred_producer:
                    return func(*args, **kwargs)
            finally:
                setattr(component, key, sns_producer)
        return wrapper
    return decorator


def collapse_dict(dct, prefix="", separator="."):
    """
    Collapse a nested dictionary into a single-level dictionary.

    Since "." is not legal in an environment variable, we can't easily express our
    pubsub overrides in environment variable form. The only legal special characters
    is underscore and our configuration loader already uses double underscore as
    a dictionary separator. So we (re)collapse nested dictionaries here.

    """
    for key, value in dct.items():
        if isinstance(value, dict):
            for nested_key, nested_value in collapse_dict(value, key, separator):
                yield separator.join([prefix, nested_key]), nested_value
        else:
            yield separator.join([prefix, key]), value


def iter_topic_mappings(dct):
    for key, value in dct.items():
        if isinstance(value, str):
            yield key, value
        else:
            for nested_key, nested_value in collapse_dict(value, key):
                yield nested_key, nested_value


@defaults(
    default=None,
    mappings={},
)
def configure_sns_topic_arns(graph):
    """
    Configure a mapping from message types to topic ARNs.

    """
    if graph.config.sns_topic_arns.default is None:
        sns_topic_arns = dict()
    else:
        sns_topic_arns = defaultdict(lambda: graph.config.sns_topic_arns.default)
        # NB: Do not use the default for the batch schema
        sns_topic_arns[MessageBatchSchema.MEDIA_TYPE] = None

    sns_topic_arns.update(graph.config.sns_topic_arns.mappings)

    for lifecycle_change in graph.pubsub_lifecycle_change:
        resource_dict = graph.config.sns_topic_arns.get(lifecycle_change, {})
        for resource_name, topic in iter_topic_mappings(resource_dict):
            media_type = make_media_type(resource_name, lifecycle_change)
            sns_topic_arns[media_type] = topic

    return sns_topic_arns


@defaults(
    profile_name=None,
    region_name=None,
    endpoint_url=None,
    mock_sns=True,
    skip=None,
    # the size used to determine batching in the deferred batch producer
    deferred_batch_size=typed(int, default_value=100),
    # SNS endpoint timeout configuration
    connect_timeout=typed(int, default_value=60),
    read_timeout=typed(int, default_value=60),
)
def configure_sns_producer(graph):
    """
    Configure an SNS producer.

    The SNS Producer requires the following collaborators:
        - Opaque from microcosm.opaque for capturing context information
        - an aws sns client, i.e. from boto.
        - pubsub message codecs: see tests for examples.
        - sns topic arns: see tests for examples.

    """
    if graph.metadata.testing:
        from unittest.mock import MagicMock

        if not graph.config.sns_producer.mock_sns:
            return MagicMock()

        sns_client = MagicMock()
    else:
        endpoint_url = graph.config.sns_producer.endpoint_url
        profile_name = graph.config.sns_producer.profile_name
        region_name = graph.config.sns_producer.region_name
        session = Session(profile_name=profile_name)
        config = Config(
            read_timeout=graph.config.sns_producer.read_timeout,
            connect_timeout=graph.config.sns_producer.connect_timeout,
        )
        sns_client = session.client(
            "sns",
            endpoint_url=endpoint_url,
            region_name=region_name,
            config=config,
        )
    try:
        opaque = graph.opaque
    except NotBoundError:
        opaque = None

    if graph.config.sns_producer.skip is None:
        # In development mode, default to not publishing because there's typically
        # not anywhere to publish to (e.g. no SNS topic)
        skip = graph.metadata.debug
    else:
        # If configured explicitly, respect the flag
        skip = strtobool(graph.config.sns_producer.skip)

    return SNSProducer(
        opaque=opaque,
        pubsub_message_schema_registry=graph.pubsub_message_schema_registry,
        sns_client=sns_client,
        sns_topic_arns=graph.sns_topic_arns,
        skip=skip,
        deferred_batch_size=graph.config.sns_producer.deferred_batch_size,
        pubsub_producer_metrics=graph.pubsub_producer_metrics,
    )



"""
A single SQS message.

"""
from microcosm_pubsub.constants import TTL_KEY


class SQSMessage:
    """
    SQS message wrapper.

    """
    def __init__(self,
                 consumer,
                 content,
                 media_type,
                 message_id,
                 receipt_handle,
                 topic_arn=None,
                 approximate_receive_count=None,
                 handler=None):
        self.consumer = consumer
        self.content = content
        self.media_type = media_type
        self.message_id = message_id
        self.receipt_handle = receipt_handle
        self.topic_arn = topic_arn
        self.approximate_receive_count = approximate_receive_count
        self.handler = handler

    def ack(self):
        """
        Acknowledge this message was processed successfully.

        """
        self.consumer.ack(self)

    def nack(self, visibility_timeout_seconds=None):
        """
        Acknowledge this message was NOT processed successfully.

        """
        self.consumer.nack(self, visibility_timeout_seconds)

    @property
    def opaque_data(self):
        if not self.content:
            return dict()
        return self.content.get("opaque_data", {})

    @property
    def ttl(self):
        if TTL_KEY not in self.opaque_data:
            return None
        return int(self.opaque_data[TTL_KEY])

    @property
    def uri(self):
        if not self.content:
            return None
        return self.content.get("uri")



#!/usr/bin/env python
from setuptools import find_packages, setup


project = "microcosm-pubsub"
version = "2.22.0"


setup(
    name=project,
    version=version,
    description="PubSub with SNS/SQS",
    author="Globality Engineering",
    author_email="engineering@globality.com",
    url="https://github.com/globality-corp/microcosm-pubsub",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.6",
    install_requires=[
        "tornado<6",
        "boto3>=1.5.8",
        "dataclasses;python_version<'3.7'",
        "marshmallow>=3.0.0",
        "microcosm>=2.12.0",
        "microcosm-caching>=0.2.0",
        "microcosm-daemon>=1.0.0",
        "microcosm-logging>=1.3.0",
    ],
    extras_require={
        "metrics": "microcosm-metrics>=2.5.0",
        "sentry": "sentry-sdk>=0.14.4",
        "test": [
            "sentry-sdk>=0.14.4",
            "PyHamcrest",
            "coverage",
            "parameterized",
        ],
    },
    setup_requires=[
        "nose>=1.3.6",
    ],
    dependency_links=[
    ],
    entry_points={
        "console_scripts": [
            "publish-naive = microcosm_pubsub.main:make_naive_message",
            "pubsub = microcosm_pubsub.main:main",
        ],
        "microcosm.factories": [
            "pubsub_message_schema_registry = microcosm_pubsub.registry:configure_schema_registry",
            "pubsub_lifecycle_change = microcosm_pubsub.conventions:LifecycleChange",
            "pubsub_send_batch_metrics = microcosm_pubsub.metrics:PubSubSendBatchMetrics",
            "pubsub_send_metrics = microcosm_pubsub.metrics:PubSubSendMetrics",
            "pubsub_producer_metrics = microcosm_pubsub.metrics:PubSubProducerMetrics",
            "sqs_message_context = microcosm_pubsub.context:SQSMessageContext",
            "sqs_consumer = microcosm_pubsub.consumer:configure_sqs_consumer",
            "sqs_envelope = microcosm_pubsub.envelope:configure_sqs_envelope",
            "sqs_message_dispatcher = microcosm_pubsub.dispatcher:SQSMessageDispatcher",
            "sqs_message_handler_registry = microcosm_pubsub.registry:configure_handler_registry",
            "sns_producer = microcosm_pubsub.producer:configure_sns_producer",
            "sns_topic_arns = microcosm_pubsub.producer:configure_sns_topic_arns",
            "sentry_logging_pubsub = microcosm_pubsub.sentry:configure_sentry_pubsub",
        ],
    },
    tests_require=[
        "coverage>=3.7.1",
        "parameterized>=0.7.0",
        "PyHamcrest>=1.8.5",
        "parameterized>=0.7.4",
    ],
)



"""
Registry of SQS message handlers.

"""
from collections import defaultdict
from inspect import isclass

from marshmallow import EXCLUDE
from microcosm.api import defaults
from microcosm_logging.decorators import logger

from microcosm_pubsub.codecs import PubSubMessageCodec
from microcosm_pubsub.conventions.messages import (
    ChangedURIMessageSchema,
    IdentityMessageSchema,
    URIMessageSchema,
)


class AlreadyRegisteredError(Exception):
    pass


@logger
class PubSubMessageSchemaRegistry:
    """
    Keeps track of available message schemas.

    """
    def __init__(self, graph):
        self._media_types = set()
        self._mappings = dict()
        self.graph = graph
        self.lifecycle_change = graph.pubsub_lifecycle_change
        self.auto_register = graph.config.pubsub_message_schema_registry.auto_register

    def register(self, media_type, value):
        """
        Register a value for a media type.

        It is an error to register more than one value for the same media type.

        """
        self._media_types.add(media_type)

        if isinstance(value, str):
            # When using the @handles convention, there may not be a concrete schema class declared.
            # In this case, the media type itself is passed as the value.
            # thereby allowing handlers to be declared using convention-driven
            return

        existing_value = self._mappings.get(media_type)
        if existing_value:
            if value == existing_value:
                return
            raise AlreadyRegisteredError("A schema already exists for media type: {}".format(
                media_type,
            ))
        self._mappings[media_type] = value

    def find_matching_media_type(self, media_type):
        """
        Find if a media type is present in the registry, either as an exact match,
        or as a pattern match

        The only non-exact pattern currently supported is a trailing *, which can match any character except `.`
        e.g. "application/vnd.globality.pubsub._.created.foo.*"

        """
        # `media_type` is present as an exact match
        if media_type in self._media_types:
            return media_type
        # Otherwise look for a pattern match
        media_type_prefix = ".".join(media_type.split(".")[0:-1])
        wildcard_media_type = f"{media_type_prefix}.*"
        if wildcard_media_type in self._media_types:
            return wildcard_media_type

        return None

    def find(self, media_type):
        """
        Create a codec or raise KeyError. If autoregistration is enabled, falls
        back to the URIMessageSchema.

        """
        matching_media_type = self.find_matching_media_type(media_type)
        if matching_media_type is None:
            if self.auto_register and self.lifecycle_change.matches(media_type):
                # When using convention-based media types, we may need to auto-register
                self._media_types.add(media_type)
            else:
                raise KeyError("Unregistered media type: {}".format(media_type))

        try:
            # use a concrete schema class if any
            schema_cls = self._mappings[matching_media_type]
            schema = schema_cls(unknown=EXCLUDE)
        except KeyError:
            # use convention otherwise
            if self.lifecycle_change.Deleted in media_type.split("."):
                schema = IdentityMessageSchema()
            elif self.lifecycle_change.Changed in media_type.split("."):
                schema = ChangedURIMessageSchema()
            else:
                schema = URIMessageSchema()

        return PubSubMessageCodec(schema)


@logger
class SQSMessageHandlerRegistry:
    """
    Keeps track of available handlers.

    """
    def __init__(self, graph):
        self._mappings = defaultdict(set)
        self.graph = graph

    def register(self, media_type, handler):
        self._mappings[media_type].add(handler)

    def compute_bound_handlers(self, bindings):
        """
        Compute which handlers are bound to the graph.

        In cases where multiple daemons exist in the same code base, it's possible that
        the registry will include handlers that aren't currently bound.

        """
        bound_components = [
            getattr(self.graph, binding)
            for binding in bindings
        ]
        bound_component_types = [
            type(bound_component)
            for bound_component in bound_components
        ]

        bound_handlers = dict()
        for media_type, handler in self.iter_handlers():
            if handler in bound_components or handler in bound_component_types:
                if bound_handlers.get(media_type):
                    raise AlreadyRegisteredError(
                        "Handler {} already registered for media type: {}".format(
                            bound_handlers.get(media_type).__name__,
                            media_type,
                        )
                    )
                bound_handlers[media_type] = handler
        return bound_handlers

    def find(self, media_type, bound_handlers):
        handler = bound_handlers[media_type]

        if isclass(handler):
            return handler(self.graph)
        else:
            return handler

    def iter_handlers(self):
        for media_type, handlers in self._mappings.items():
            for handler in handlers:
                yield media_type, handler

    def keys(self):
        return self._mappings.keys()


@defaults(
    auto_register=True,
)
def configure_schema_registry(graph):
    return PubSubMessageSchemaRegistry(graph)


def configure_handler_registry(graph):
    return SQSMessageHandlerRegistry(graph)


def media_type_for(schema_cls):
    if hasattr(schema_cls, "MEDIA_TYPE"):
        return schema_cls.MEDIA_TYPE
    if hasattr(schema_cls, "infer_media_type"):
        return schema_cls.infer_media_type()
    if isinstance(schema_cls, str):
        return schema_cls
    raise Exception("Cannot infer media type for schema class: {}".format(schema_cls))
